import click

from stela.cmd import stela_init


def print_title(title: str):
    title_length = len(title)
    box_width = title_length + 4

    click.echo("*" * box_width)
    click.echo("* " + " " * title_length + " *")
    click.echo("* " + title + " *")
    click.echo("* " + " " * title_length + " *")
    click.echo("*" * box_width)


@click.group("cli")
@click.pass_context
def cli(ctx):
    # No need for code, just group commands.
    pass


@cli.command()
@click.option("--default", is_flag=True, help="Use Default Values.")
@click.option("--no-confirm", is_flag=True, help="Do not ask for confirmation.")
def init(default, no_confirm):
    """Initialize Stela for your project."""
    print_title("Initializing Stela")
    if not no_confirm:
        click.pause(
            "This command will configure stela for your project.\nPress a key to continue."
        )
    initializer = stela_init.StelaInit(".")
    initializer.run(default)


@cli.command()
def gen_stub():
    """Generate Stela Stub."""
    from stela.config import StelaOptions
    from stela.helpers.stub import create_stela_stub
    from stela.main import StelaMain

    stela_config = StelaOptions.get_config()
    stela_data = StelaMain(options=stela_config)
    stela_data.get_project_settings()

    print_title("Generate Stela Stub file")
    result, message = create_stela_stub(stela_data.settings)
    click.secho(
        f"{'Success' if result else 'Error'}: {message}",
        fg="green" if result else "red",
    )


if __name__ == "__main__":
    cli()



"""Stela Utils module."""

from enum import Enum, unique
from importlib import reload
from pathlib import Path
from typing import Any, Dict, Optional


@unique
class StelaFileType(Enum):
    """Available formats for configuration files."""

    INI = [".ini"]
    JSON = [".json"]
    YAML = [".yaml", ".yml"]
    TOML = [".toml"]
    ENV = [".env"]


def read_env() -> "StelaDot":  # type: ignore
    """Reload Stela configuration.

    This helper will import again stela module
    to update settings.

    """
    import stela

    reload(stela)
    return stela.env


def find_file_folder(file_name: str) -> Optional[Path]:
    """Find base folder for named file.

    For pyproject.toml and conf_stela.py files.

    Look for:
        1. Current working directory up to root
        2. Current file directory up to root


    :return: Optional Path
    """

    def look_for_file(current_path: Path) -> Optional[Path]:
        if current_path.joinpath(file_name).exists():
            return current_path
        if str(current_path) in ["/", "\\"] or current_path.parent == current_path:
            return None
        return look_for_file(current_path.parent)

    found_path = look_for_file(Path().cwd())
    if found_path:
        return found_path

    found_path = look_for_file(Path(__file__))
    return found_path


def merge_dicts(source_dict: Dict[Any, Any], target_dict: Dict[Any, Any]) -> None:
    """Merge source dictionary into target.

    Example:
        # from pyproject.toml

        [environment]
        test = true
        project.foo = "bar"
        project.x = "a"

        [environment.local]
        project.bar = "foo"
        project.x = "b"
        project.db.name = "test"

    Expected data for ENVIRONMENT=local:
        settings.to_dict = {
            "test": True,
            "project": {
                "foo": "bar,
                "project": {
                    "x": "b"
                },
                "bar": "foo",
                "db": {
                    "name": "test"
                },
            }
        }
    :param source_dict: The Source dictionary
    :param target_dict: The Target dictionary, which will me merged
    :current_key: Current Source Dictionary Key
    """

    def _merge_dicts(source, target, current_key):
        if isinstance(source[current_key], dict):
            if not target.get(current_key):
                target[current_key] = {}
            for k in source[current_key]:
                if not target[current_key].get(k):
                    target[current_key][k] = {}
                _merge_dicts(source[current_key], target[current_key], k)
        else:
            target[current_key] = source[current_key]

    for key in source_dict.keys():
        _merge_dicts(source_dict, target_dict, key)


def show_value(value: str, filter_values: bool) -> str:
    """Shows dictionary value as plain text or filtered.

    For filtered values with less than 3 characters
    *** will be printed.

    """
    value = str(value)
    if filter_values:
        if len(value) < 4:
            return "***"
        visible_chars = min(int(len(value) / 3), 3)
        value = f"{value[:visible_chars]}***{value[-visible_chars:]}"
    return value


def get_base_path():
    path = find_file_folder(".stela")
    if not path:
        path = find_file_folder("pyproject.toml")
    if not path:
        path = Path().cwd()
    return path


def get_stela_config():
    """Return Stela Options instance.

    :return StelaOptions:
    """
    from stela.config import StelaOptions

    config = StelaOptions.get_config()
    return config


def read_env_files() -> Dict[str, Any]:
    """Return data parsed from env files.

    :return Dict[str, Any]:
    """
    from stela.main import StelaMain

    config = get_stela_config()
    data = StelaMain(config)
    data.read_env_files()
    return data.settings


def flatten_dict(d, parent_key="", sep="_"):
    """Flatten dict.

    :param d: Dict[Any, Any]
    :param parent_key: str
    :param sep: str
    :returns: Dict[str, Any]
    """
    items = []
    for k, v in d.items():
        new_key = f"{parent_key}{sep}{k}" if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key.upper(), v))
    return dict(items)


def merge_env(
    target: dict, env_data=dict[str, Any], separator: str = "_"
) -> dict[str, Any]:
    """Merge Env Data in target dict.

    Target dict can be a nested dictionary, with lower case keys.

    Example::
        {
          "project": {
             "debug": False
          },
        }

    Env data is always a flatten dict, with upper case keys.

    Example::
        {
          "PROJECT_DEBUG": True,
        }

    We need to flatten target dict merging keys using informed separator.
    After that, we will update target dict using the Env data.

    :param target: Dict[Any, Any]: dict to be merged
    :param env_data: Dict[Str, Any]: dict with env data
    :param separator: str: separator to flatten dict keys
    :returns: Dict[str, Any]
    """
    flat_dict = flatten_dict(target, sep=separator)
    flat_dict.update(env_data)

    return flat_dict



import os
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

from loguru import logger

from stela.exceptions import StelaEnvironmentNotFoundError, StelaFileTypeError
from stela.parsers.dotenv import read_dotenv
from stela.parsers.reader import StelaFileReader
from stela.utils import StelaFileType, find_file_folder


@dataclass
class StelaOptions:
    config_file_extension: StelaFileType = StelaFileType.INI
    env_file: str = ".env"
    log_filtered_value: bool = True
    show_logs: bool = False
    config_file_path: str = "."
    warn_if_env_is_missing: bool = False
    dotenv_encoding: str = "utf-8"
    _dotenv_data: Optional[Dict[Any, Any]] = None
    current_environment: Optional[str] = None
    env_table: str = "environment"
    _filenames: List[str] = field(default_factory=list)
    environment_variable_name: str = "STELA_ENV"
    final_loader: str = "stela.main.default_loader"
    raise_on_missing_variable: bool = True
    evaluate_data: bool = True
    default_environment: Optional[str] = None
    no_env_name: str = "GLOBAL"

    @classmethod
    def get_from_env_or_settings(
        cls, key: str, file_settings: Dict[str, Any], default: Any
    ) -> Any:
        """Get from environment or settings the Stela Option argument.

        :param key: Current stela argument.
        :param file_settings: Stela Data from pyproject.toml.
        :param default: default value for argument.
        :return: Any
        """
        normalized_key = key.upper().replace(".", "_")
        if key.startswith("STELA_"):
            environment_name = normalized_key
        else:
            environment_name = f"STELA_{normalized_key}"
        value = os.getenv(environment_name, file_settings.get(key, default))
        if value and isinstance(value, str) and isinstance(default, bool):
            return value.lower() == "true"
        return value

    @classmethod
    def _get_current_environment(cls, settings):
        # Get from Memory
        settings["current_environment"] = os.getenv(
            settings["environment_variable_name"]
        )

        # Get from .env if available
        if not settings["current_environment"]:
            settings["current_environment"] = settings["_dotenv_data"].get(
                settings["environment_variable_name"]
            )

        # Get from Default
        if not settings["current_environment"]:
            settings["current_environment"] = settings["default_environment"]

        # No environment found.
        if not settings["current_environment"] and settings.get(
            "use_environment_layers"
        ):
            raise StelaEnvironmentNotFoundError("Environment not found.")

    @classmethod
    def _get_dotenv_data(cls, settings):
        """Get data from dotenv files."""
        settings["_dotenv_data"] = {}
        if not settings.get("do_not_read_dotenv"):
            files = [
                settings["env_file"],
                f"{settings['env_file']}.local",
            ]
            if (
                settings.get("default_environment")
                and settings["default_environment"] != "GLOBAL"
            ):
                files += [
                    f"{settings['env_file']}.{settings['default_environment'].lower()}",
                    f"{settings['env_file']}.{settings['default_environment'].lower()}.local",
                ]
            for file in files:
                settings["_dotenv_data"] |= read_dotenv(
                    config_file_path=settings["config_file_path"],
                    env_file=file,
                    encoding=settings["dotenv_encoding"],
                    verbose=settings["warn_if_env_is_missing"],
                    show_logs=settings["show_logs"],
                    filter_logs=settings["log_filtered_value"],
                )
            # Filter dict for keys which start with "STELA_"
            settings["_dotenv_data"] = {
                k: v
                for k, v in settings["_dotenv_data"].items()
                if k.startswith("STELA_")
            }

    def get_extensions(self) -> List[str]:
        """Return file extensions for project configuration files."""
        return self.config_file_extension.value  # type: ignore

    @classmethod
    def get_config(cls) -> "StelaOptions":
        """Get config from pyproject.toml."""

        file_settings, settings = cls.get_settings()
        try:
            config_file_extension = cls.get_from_env_or_settings(
                "config_file_extension",
                file_settings,
                cls.config_file_extension.value[0].replace(".", "").upper(),
            )
            settings["config_file_extension"] = StelaFileType[config_file_extension]
        except KeyError:
            raise StelaFileTypeError(
                f"Invalid file type: {file_settings.get('config_file_extension')}"
            )
        cls._get_dotenv_data(settings)
        cls._get_current_environment(settings)
        settings["_filenames"] = [
            f"{settings.get('config_file_prefix', '')}{settings['current_environment']}{extension}"
            for extension in settings["config_file_extension"].value
        ]
        return cls(**settings)

    @classmethod
    def get_settings(cls):
        file_settings = {}
        reader = StelaFileReader()
        toml_path = find_file_folder("pyproject.toml")
        if toml_path:
            toml_settings = reader.load_toml(toml_path.joinpath("pyproject.toml"))
            file_settings = toml_settings.get("tool", {}).get("stela", {})
            if file_settings:
                logger.info("Using pyproject.toml for stela settings.")
        if not file_settings:
            ini_path = find_file_folder(".stela")
            if ini_path:
                ini_settings = reader.load_ini(ini_path.joinpath(".stela"))
                file_settings = ini_settings.get("stela", {})
                if file_settings:
                    logger.info("Using .stela for stela settings.")
        if not file_settings:
            logger.info(
                "No stela file configuration found. Using default stela settings."
            )
        settings = {
            "environment_variable_name": cls.get_from_env_or_settings(
                "environment_variable_name",
                file_settings,
                cls.environment_variable_name,
            ),
            "default_environment": cls.get_from_env_or_settings(
                "default_environment", file_settings, cls.default_environment
            ),
            "evaluate_data": cls.get_from_env_or_settings(
                "evaluate_data", file_settings, cls.evaluate_data
            ),
            "config_file_path": cls.get_from_env_or_settings(
                "config_file_path", file_settings, cls.config_file_path
            ),
            "show_logs": cls.get_from_env_or_settings(
                "show_logs", file_settings, cls.show_logs
            ),
            "log_filtered_value": cls.get_from_env_or_settings(
                "log_filtered_value",
                file_settings,
                cls.log_filtered_value,
            ),
            "env_file": cls.get_from_env_or_settings(
                "env_file", file_settings, cls.env_file
            ),
            "warn_if_env_is_missing": cls.get_from_env_or_settings(
                "warn_if_env_is_missing", file_settings, cls.warn_if_env_is_missing
            ),
            "dotenv_encoding": cls.get_from_env_or_settings(
                "dotenv_encoding", file_settings, cls.dotenv_encoding
            ),
            "final_loader": cls.get_from_env_or_settings(
                "final_loader", file_settings, cls.final_loader
            ),
            "raise_on_missing_variable": cls.get_from_env_or_settings(
                "raise_on_missing_variable",
                file_settings,
                cls.raise_on_missing_variable,
            ),
            "no_env_name": cls.get_from_env_or_settings(
                "no_env_name", file_settings, cls.no_env_name
            ),
        }
        return file_settings, settings



"""Stela Decorators."""

from functools import wraps


def stela_enable_logs(f):
    from loguru import logger

    from stela import env

    @wraps(f)
    def wrapper(*args, **kwargs):
        logger.enable("stela")
        ret = f(*args, **kwargs)
        logger.disable("stela")
        if env._stela_options.show_logs:
            logger.enable("stela")
        return ret

    return wrapper


def stela_disable_logs(f):
    from loguru import logger

    from stela import env

    @wraps(f)
    def wrapper(*args, **kwargs):
        logger.disable("stela")
        ret = f(*args, **kwargs)
        if env._stela_options.show_logs:
            logger.enable("stela")
        return ret

    return wrapper
